from __future__ import print_function

import json
import logging

from .stream import (
    MarketStream,
    OrderStream,
)

logger = logging.getLogger(__name__)


class BaseListener(object):

    def __init__(self, max_latency=0.5):
        self.max_latency = max_latency
        self.market_stream = None
        self.order_stream = None

    def register_stream(self, unique_id, operation):
        if operation == 'authentication':
            logger.info('[Listener: %s]: %s' % (unique_id, operation))

        elif operation == 'marketSubscription':
            if self.market_stream is not None:
                logger.warning('[Listener: %s]: marketSubscription stream already registered, replacing data' %
                               unique_id)
            self.market_stream = self._add_stream(unique_id, operation)

        elif operation == 'orderSubscription':
            if self.order_stream is not None:
                logger.warning('[Listener: %s]: orderSubscription stream already registered, replacing data' %
                               unique_id)
            self.order_stream = self._add_stream(unique_id, operation)

    def on_data(self, raw_data):
        print(raw_data)

    def _add_stream(self, unique_id, operation):
        print('Register: %s %s' % (operation, unique_id))

    def __str__(self):
        return '<BaseListener>'

    def __repr__(self):
        return str(self)


class StreamListener(BaseListener):
    """Stream listener, processes results from socket,
    holds a market and order stream which hold
    market_book caches
    """

    def __init__(self, output_queue, max_latency=0.5, lightweight=False):
        super(StreamListener, self).__init__(max_latency)
        self.output_queue = output_queue
        self.lightweight = lightweight

    def on_data(self, raw_data):
        """Called when raw data is received from connection.
        Override this method if you wish to manually handle
        the stream data

        :param raw_data: Received raw data
        :return: Return False to stop stream and close connection
        """
        try:
            data = json.loads(raw_data)
        except ValueError:
            logger.error('value error: %s' % raw_data)
            return
        unique_id = data.get('id')

        if self._error_handler(data, unique_id):
            return False

        operation = data.get('op')
        if operation == 'connection':
            self._on_connection(data, unique_id)
        elif operation == 'status':
            self._on_status(data, unique_id)
        elif operation in ['mcm', 'ocm']:
            self._on_change_message(data, unique_id)

    def _on_connection(self, data, unique_id):
        """Called on collection operation

        :param data: Received data
        """
        self.connection_id = data.get('connectionId')
        logger.info('[Connect: %s]: connection_id: %s' % (unique_id, self.connection_id))

    @staticmethod
    def _on_status(data, unique_id):
        """Called on status operation

        :param data: Received data
        """
        status_code = data.get('statusCode')
        logger.info('[Subscription: %s]: %s' % (unique_id, status_code))

    def _on_change_message(self, data, unique_id):
        change_type = data.get('ct', 'UPDATE')
        operation = data.get('op')

        if operation == 'mcm':
            stream = self.market_stream
        else:
            stream = self.order_stream

        logger.debug('[Subscription: %s]: %s: %s' % (unique_id, change_type, data))

        if change_type == 'SUB_IMAGE':
            stream.on_subscribe(data)
        elif change_type == 'RESUB_DELTA':
            stream.on_resubscribe(data)
        elif change_type == 'HEARTBEAT':
            stream.on_heartbeat(data)
        elif change_type == 'UPDATE':
            stream.on_update(data)

    def _add_stream(self, unique_id, stream_type):
        if stream_type == 'marketSubscription':
            return MarketStream(
                unique_id, self.output_queue, self.max_latency, self.lightweight
            )
        elif stream_type == 'orderSubscription':
            return OrderStream(
                unique_id, self.output_queue, self.max_latency, self.lightweight
            )

    @staticmethod
    def _error_handler(data, unique_id):
        """Called when data first received

        :param data: Received data
        :param unique_id: Unique id
        :return: True if error present
        """
        if data.get('statusCode') == 'FAILURE':
            logger.error('[Subscription: %s] %s: %s' % (unique_id, data.get('errorCode'), data.get('errorMessage')))
            if data.get('connectionClosed'):
                return True

    def __str__(self):
        return 'StreamListener'

    def __repr__(self):
        return '<StreamListener>'



import datetime
from operator import itemgetter

from ..utils import update_available
from .baseresource import BaseResource
from .bettingresources import (
    MarketBook,
    CurrentOrders,
)
from ..enums import (
    StreamingOrderType,
    StreamingPersistenceType,
    StreamingSide,
    StreamingStatus,
)


class MarketDefinitionRunner(object):
    """
    :type adjustment_factor: float
    :type id: int
    :type removal_date: datetime.datetime
    :type sort_priority: int
    :type status: unicode
    """

    def __init__(self, id, sortPriority, status, hc=None, bsp=None, adjustmentFactor=None, removalDate=None):
        self.selection_id = id
        self.sort_priority = sortPriority
        self.status = status
        self.handicap = hc
        self.bsp = bsp
        self.adjustment_factor = adjustmentFactor
        self.removal_date = BaseResource.strip_datetime(removalDate)

    def __str__(self):
        return 'MarketDefinitionRunner: %s' % self.selection_id

    def __repr__(self):
        return '<MarketDefinitionRunner>'


class MarketDefinition(object):
    """
    :type bet_delay: int
    :type betting_type: unicode
    :type bsp_market: bool
    :type bsp_reconciled: bool
    :type complete: bool
    :type country_code: unicode
    :type cross_matching: bool
    :type discount_allowed: bool
    :type event_id: unicode
    :type event_type_id: unicode
    :type in_play: bool
    :type market_base_rate: int
    :type market_time: datetime.datetime
    :type market_type: unicode
    :type number_of_active_runners: int
    :type number_of_winners: int
    :type open_date: datetime.datetime
    :type persistence_enabled: bool
    :type regulators: list[unicode]
    :type runners: list[MarketDefinitionRunner]
    :type runners_voidable: bool
    :type settled_time: datetime.datetime
    :type status: unicode
    :type suspend_time: datetime.datetime
    :type timezone: unicode
    :type turn_in_play_enabled: bool
    :type venue: unicode
    :type version: int
    """

    def __init__(self, betDelay, bettingType, bspMarket, bspReconciled, complete, crossMatching, discountAllowed,
                 eventId, eventTypeId, inPlay, marketBaseRate, marketTime, numberOfActiveRunners, numberOfWinners,
                 openDate, persistenceEnabled, regulators, runnersVoidable, status, timezone, turnInPlayEnabled,
                 version, runners, countryCode=None, eachWayDivisor=None, venue=None, settledTime=None,
                 suspendTime=None, marketType=None, lineMaxUnit=None, lineMinUnit=None, lineInterval=None):
        self.bet_delay = betDelay
        self.betting_type = bettingType
        self.bsp_market = bspMarket
        self.bsp_reconciled = bspReconciled
        self.complete = complete
        self.country_code = countryCode
        self.cross_matching = crossMatching
        self.discount_allowed = discountAllowed
        self.event_id = eventId
        self.event_type_id = eventTypeId
        self.in_play = inPlay
        self.market_base_rate = marketBaseRate
        self.market_time = BaseResource.strip_datetime(marketTime)
        self.market_type = marketType
        self.number_of_active_runners = numberOfActiveRunners
        self.number_of_winners = numberOfWinners
        self.open_date = BaseResource.strip_datetime(openDate)
        self.persistence_enabled = persistenceEnabled
        self.regulators = regulators
        self.runners_voidable = runnersVoidable
        self.settled_time = BaseResource.strip_datetime(settledTime)
        self.status = status
        self.each_way_divisor = eachWayDivisor
        self.suspend_time = BaseResource.strip_datetime(suspendTime)
        self.timezone = timezone
        self.turn_in_play_enabled = turnInPlayEnabled
        self.venue = venue
        self.version = version
        self.line_max_unit = lineMaxUnit
        self.line_min_unit = lineMinUnit
        self.line_interval = lineInterval
        self.runners = [MarketDefinitionRunner(**i) for i in runners]


class RunnerBook(object):

    def __init__(self, id, ltp=None, tv=None, trd=None, atb=None, batb=None, bdatb=None, atl=None, batl=None,
                 bdatl=None, spn=None, spf=None, spb=None, spl=None, hc=None):
        self.selection_id = id
        self.last_price_traded = ltp
        self.total_matched = tv
        self.traded = trd
        self.available_to_back = atb
        self.best_available_to_back = batb
        self.best_display_available_to_back = bdatb
        self.available_to_lay = atl
        self.best_available_to_lay = batl
        self.best_display_available_to_lay = bdatl
        self.starting_price_near = spn
        self.starting_price_far = spf
        self.starting_price_back = spb
        self.starting_price_lay = spl
        self.handicap = hc

    def update_traded(self, traded_update):
        """:param traded_update: [price, size]
        """
        if not traded_update:
            self.traded = traded_update
        elif self.traded is None:
            self.traded = traded_update
        else:
            update_available(self.traded, traded_update, 1)

    def update_available_to_back(self, book_update):
        """:param book_update: [price, size]
        """
        if self.available_to_back is None:
            self.available_to_back = book_update
        else:
            update_available(self.available_to_back, book_update, 1)

    def update_available_to_lay(self, book_update):
        """:param book_update: [price, size]
        """
        if self.available_to_lay is None:
            self.available_to_lay = book_update
        else:
            update_available(self.available_to_lay, book_update, 1)

    def update_best_available_to_back(self, book_update):
        """:param book_update: [level, price, size]
        """
        if self.best_available_to_back is None:
            self.best_available_to_back = book_update
        else:
            update_available(self.best_available_to_back, book_update, 2)

    def update_best_available_to_lay(self, book_update):
        """:param book_update: [level, price, size]
        """
        if self.best_available_to_lay is None:
            self.best_available_to_lay = book_update
        else:
            update_available(self.best_available_to_lay, book_update, 2)

    def update_best_display_available_to_back(self, book_update):
        """:param book_update: [level, price, size]
        """
        if self.best_display_available_to_back is None:
            self.best_display_available_to_back = book_update
        else:
            update_available(self.best_display_available_to_back, book_update, 2)

    def update_best_display_available_to_lay(self, book_update):
        """:param book_update: [level, price, size]
        """
        if self.best_display_available_to_lay is None:
            self.best_display_available_to_lay = book_update
        else:
            update_available(self.best_display_available_to_lay, book_update, 2)

    def update_starting_price_back(self, book_update):
        """:param book_update: [price, size]
        """
        if self.starting_price_back is None:
            self.starting_price_back = book_update
        else:
            update_available(self.starting_price_back, book_update, 1)

    def update_starting_price_lay(self, book_update):
        """:param book_update: [price, size]
        """
        if self.starting_price_lay is None:
            self.starting_price_lay = book_update
        else:
            update_available(self.starting_price_lay, book_update, 1)

    @property
    def serialise_traded_volume(self):
        if self.traded:
            return [
                {'price': volume[0], 'size': volume[1]} for volume in sorted(self.traded, key=itemgetter(0))
            ]
        else:
            return []

    @property
    def serialise_available_to_back(self):
        if self.available_to_back:
            return [{'price': volume[0], 'size': volume[1]}
                    for volume in sorted(self.available_to_back, key=itemgetter(0), reverse=True)]
        elif self.best_display_available_to_back:
            return [{'price': volume[1], 'size': volume[2]}
                    for volume in sorted(self.best_display_available_to_back, key=itemgetter(0))]
        elif self.best_available_to_back:
            return [{'price': volume[1], 'size': volume[2]}
                    for volume in sorted(self.best_available_to_back, key=itemgetter(0))]
        else:
            return []

    @property
    def serialise_available_to_lay(self):
        if self.available_to_lay:
            return [{'price': volume[0], 'size': volume[1]}
                    for volume in sorted(self.available_to_lay, key=itemgetter(0))]
        elif self.best_display_available_to_lay:
            return [{'price': volume[1], 'size': volume[2]}
                    for volume in sorted(self.best_display_available_to_lay, key=itemgetter(0))]
        elif self.best_available_to_lay:
            return [{'price': volume[1], 'size': volume[2]}
                    for volume in sorted(self.best_available_to_lay, key=itemgetter(0))]
        return []

    @property
    def serialise_starting_price_back(self):
        if self.starting_price_back:
            return [{'price': volume[0], 'size': volume[1]}
                    for volume in sorted(self.starting_price_back, key=itemgetter(0))]
        return []

    @property
    def serialise_starting_price_lay(self):
        if self.starting_price_lay:
            return [{'price': volume[0], 'size': volume[1]}
                    for volume in sorted(self.starting_price_lay, key=itemgetter(0))]
        return []

    def serialise(self, runner_definition):
        return {
            'status': runner_definition.status,
            'ex': {
                'tradedVolume': self.serialise_traded_volume,
                'availableToBack': self.serialise_available_to_back,
                'availableToLay': self.serialise_available_to_lay
            },
            'sp': {
                'nearPrice': self.starting_price_near,
                'farPrice': self.starting_price_far,
                'backStakeTaken': self.serialise_starting_price_back,
                'layLiabilityTaken': self.serialise_starting_price_lay,
                'actualSP': runner_definition.bsp
            },
            'adjustmentFactor': runner_definition.adjustment_factor,
            'removalDate': runner_definition.removal_date,
            'lastPriceTraded': self.last_price_traded,
            'handicap': self.handicap,
            'totalMatched': self.total_matched,
            'selectionId': self.selection_id
        }


class MarketBookCache(BaseResource):

    def __init__(self, **kwargs):
        super(MarketBookCache, self).__init__(**kwargs)
        self.publish_time = kwargs.get('publish_time')
        self.market_id = kwargs.get('id')
        self.image = kwargs.get('img')
        self.total_matched = kwargs.get('tv')
        self.market_definition = MarketDefinition(**kwargs['marketDefinition']) if \
            'marketDefinition' in kwargs else None
        self.runners = [RunnerBook(**i) for i in kwargs.get('rc', [])]

    def update_cache(self, market_change, publish_time):
        self._datetime_updated = self.strip_datetime(publish_time)
        self.publish_time = publish_time

        if 'marketDefinition' in market_change:
            self.market_definition = MarketDefinition(**market_change['marketDefinition'])

        if 'tv' in market_change:
            self.total_matched = market_change['tv']

        if 'rc' in market_change:
            for new_data in market_change['rc']:
                selection_id = new_data['id']
                runner = self.runner_dict.get(selection_id)
                if runner:
                    if 'ltp' in new_data:
                        runner.last_price_traded = new_data['ltp']
                    if 'tv' in new_data:  # if runner removed tv: 0 is returned
                        runner.total_matched = new_data['tv']
                    if 'spn' in new_data:
                        runner.starting_price_near = new_data['spn']
                    if 'spf' in new_data:
                        runner.starting_price_far = new_data['spf']
                    if 'trd' in new_data:
                        runner.update_traded(new_data['trd'])
                    if 'atb' in new_data:
                        runner.update_available_to_back(new_data['atb'])
                    if 'atl' in new_data:
                        runner.update_available_to_lay(new_data['atl'])
                    if 'batb' in new_data:
                        runner.update_best_available_to_back(new_data['batb'])
                    if 'batl' in new_data:
                        runner.update_best_available_to_lay(new_data['batl'])
                    if 'bdatb' in new_data:
                        runner.update_best_display_available_to_back(new_data['bdatb'])
                    if 'bdatl' in new_data:
                        runner.update_best_display_available_to_lay(new_data['bdatl'])
                    if 'spb' in new_data:
                        runner.update_starting_price_back(new_data['spb'])
                    if 'spl' in new_data:
                        runner.update_starting_price_lay(new_data['spl'])
                else:
                    self.runners.append(RunnerBook(**new_data))

    def create_market_book(self, unique_id, streaming_update, lightweight):
        if lightweight:
            return self.serialise
        else:
            return MarketBook(
                elapsed_time=(datetime.datetime.utcnow()-self._datetime_updated).total_seconds(),
                streaming_unique_id=unique_id,
                streaming_update=streaming_update,
                market_definition=self.market_definition,
                **self.serialise
            )

    @property
    def runner_dict(self):
        return {runner.selection_id: runner for runner in self.runners}

    @property
    def market_definition_dict(self):
        return {runner.selection_id: runner for runner in self.market_definition.runners}

    @property
    def serialise(self):
        """Creates standard market book json response,
        will error if EX_MARKET_DEF not incl.
        """
        return {
            'marketId': self.market_id,
            'totalAvailable': None,
            'isMarketDataDelayed': None,
            'lastMatchTime': None,
            'betDelay': self.market_definition.bet_delay,
            'version': self.market_definition.version,
            'complete': self.market_definition.complete,
            'runnersVoidable': self.market_definition.runners_voidable,
            'totalMatched': self.total_matched,
            'status': self.market_definition.status,
            'bspReconciled': self.market_definition.bsp_reconciled,
            'crossMatching': self.market_definition.cross_matching,
            'inplay': self.market_definition.in_play,
            'numberOfWinners': self.market_definition.number_of_winners,
            'numberOfRunners': len(self.market_definition.runners),
            'numberOfActiveRunners': self.market_definition.number_of_active_runners,
            'runners': [
                runner.serialise(self.market_definition_dict.get(runner.selection_id)) for runner in self.runners
            ],
            'publishTime': self.publish_time,
        }


class UnmatchedOrder(object):

    def __init__(self, id, p, s, side, status, pt, ot, pd, sm, sr, sl, sc, sv, rac, rc, rfo, rfs,
                 md=None, avp=None, bsp=None, ld=None):
        self.bet_id = id
        self.price = p
        self.size = s
        self.bsp_liability = bsp
        self.side = side
        self.status = status
        self.persistence_type = pt
        self.order_type = ot
        self.placed_date = BaseResource.strip_datetime(pd)
        self.matched_date = BaseResource.strip_datetime(md)
        self.average_price_matched = avp
        self.size_matched = sm
        self.size_remaining = sr
        self.size_lapsed = sl
        self.size_cancelled = sc
        self.size_voided = sv
        self.regulator_auth_code = rac
        self.regulator_code = rc
        self.reference_order = rfo
        self.reference_strategy = rfs
        self.lapsed_date = ld

    def serialise(self, market_id, selection_id):
        return {
            'averagePriceMatched': self.average_price_matched or 0.0,
            'betId': self.bet_id,
            'bspLiability': self.bsp_liability,
            'handicap': 0.0,
            'marketId': market_id,
            'matchedDate': self.matched_date.strftime(
                '%Y-%m-%dT%H:%M:%S.%fZ') if self.matched_date is not None else self.matched_date,
            'orderType': StreamingOrderType[self.order_type].value,
            'persistenceType': StreamingPersistenceType[self.persistence_type].value,
            'placedDate': self.placed_date.strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
            'priceSize': {
                'price': self.price,
                'size': self.size
            },
            'regulatorCode': self.regulator_code,
            'selectionId': selection_id,
            'side': StreamingSide[self.side].value,
            'sizeCancelled': self.size_cancelled,
            'sizeLapsed': self.size_lapsed,
            'sizeMatched': self.size_matched,
            'sizeRemaining': self.size_remaining,
            'sizeVoided': self.size_voided,
            'status': StreamingStatus[self.status].value,
            'customerStrategyRef': self.reference_strategy,
            'customerOrderRef': self.reference_order,
        }


class OrderBookRunner(object):

    def __init__(self, id, fullImage=None, ml=None, mb=None, uo=None):
        self.selection_id = id
        self.full_image = fullImage
        self.matched_lays = ml
        self.matched_backs = mb
        self.unmatched_orders = [UnmatchedOrder(**i) for i in uo] if uo else []

    def update_matched_backs(self, matched_backs):
        if not self.matched_backs:
            self.matched_backs = [matched_back for matched_back in matched_backs]
        else:
            update_available(self.matched_backs, matched_backs, 1)

    def update_matched_lays(self, matched_lays):
        if not self.matched_lays:
            self.matched_lays = [matched_lay for matched_lay in matched_lays]
        else:
            update_available(self.matched_lays, matched_lays, 1)

    def update_unmatched(self, unmatched_orders):
        order_dict = {order.bet_id: order for order in self.unmatched_orders}
        for unmatched_order in unmatched_orders:
            if unmatched_order.get('id') in order_dict:
                for n, order in enumerate(self.unmatched_orders):
                    if order.bet_id == unmatched_order.get('id'):
                        self.unmatched_orders[n] = UnmatchedOrder(**unmatched_order)
                        break
            else:
                self.unmatched_orders.append(UnmatchedOrder(**unmatched_order))

    def serialise_orders(self, market_id):
        return [
            order.serialise(market_id, self.selection_id) for order in self.unmatched_orders
        ]


class OrderBookCache(BaseResource):

    def __init__(self, **kwargs):
        super(OrderBookCache, self).__init__(**kwargs)
        self.market_id = kwargs.get('id')
        self.closed = kwargs.get('closed')
        self.runners = [OrderBookRunner(**i) for i in kwargs.get('orc', [])]

    def update_cache(self, order_book, publish_time):
        self._datetime_updated = self.strip_datetime(publish_time)

        for order_changes in order_book.get('orc', []):
            selection_id = order_changes['id']
            runner = self.runner_dict.get(selection_id)
            if runner:
                runner.update_matched_lays(order_changes.get('ml', []))
                runner.update_matched_backs(order_changes.get('mb', []))
                runner.update_unmatched(order_changes.get('uo', []))
            else:
                self.runners.append(OrderBookRunner(**order_changes))

    def create_order_book(self, unique_id, streaming_update, lightweight):
        if lightweight:
            return self.serialise
        else:
            return CurrentOrders(
                elapsed_time=(datetime.datetime.utcnow()-self._datetime_updated).total_seconds(),
                streaming_unique_id=unique_id,
                streaming_update=streaming_update,
                **self.serialise
            )

    @property
    def runner_dict(self):
        return {runner.selection_id: runner for runner in self.runners}

    @property
    def serialise(self):
        orders = []
        for runner in self.runners:
            orders.extend(runner.serialise_orders(self.market_id))
        return {
            'currentOrders': orders,
            'moreAvailable': False
        }



from .baseresource import BaseResource


class Innings(object):

    def __init__(self, overs, runs, wickets):
        self.overs = overs
        self.runs = runs
        self.wickets = wickets


class HomeAwayBase(object):

    def __init__(self, penaltiesScore, penaltiesSequence, halfTimeScore, fullTimeScore, name, score=None, sets=None,
                 games=None, numberOfCards=None, numberOfCorners=None, numberOfCornersFirstHalf=None,
                 numberOfCornersSecondHalf=None, numberOfRedCards=None, numberOfYellowCards=None, highlight=None,
                 aces=None, doubleFaults=None, gameSequence=None, bookingPoints=None, isServing=None, playerSeed=None,
                 serviceBreaks=None, inning1=None):
        self.booking_points = bookingPoints
        self.full_time_score = fullTimeScore
        self.games = games
        self.half_time_score = halfTimeScore
        self.name = name
        self.number_of_cards = numberOfCards
        self.number_of_corners = numberOfCorners
        self.number_of_corners_first_half = numberOfCornersFirstHalf
        self.number_of_corners_second_half = numberOfCornersSecondHalf
        self.number_of_red_cards = numberOfRedCards
        self.number_of_yellow_cards = numberOfYellowCards
        self.penalties_score = penaltiesScore
        self.penalties_sequence = penaltiesSequence
        self.score = score
        self.sets = sets
        self.highlight = highlight
        self.aces = aces
        self.double_faults = doubleFaults
        self.game_sequence = gameSequence
        self.is_serving = isServing
        self.player_seed = playerSeed
        self.service_breaks = serviceBreaks
        self.inning1 = Innings(**inning1) if inning1 else None  # todo more innings?


class Score(object):

    def __init__(self, home, away, bookingPoints=None, numberOfCards=None, numberOfCorners=None,
                 numberOfCornersFirstHalf=None, numberOfCornersSecondHalf=None, numberOfRedCards=None,
                 numberOfYellowCards=None):
        self.booking_points = bookingPoints
        self.number_of_cards = numberOfCards
        self.number_of_corners = numberOfCorners
        self.number_of_corners_first_half = numberOfCornersFirstHalf
        self.number_of_corners_second_half = numberOfCornersSecondHalf
        self.number_of_red_cards = numberOfRedCards
        self.number_of_yellow_cards = numberOfYellowCards
        self.home = HomeAwayBase(**home)
        self.away = HomeAwayBase(**away)


class UpdateDetail(object):

    def __init__(self, elapsedRegularTime, matchTime, type, updateId, updateTime, updateType, team=None, teamName=None,
                 elapsedAddedTime=None):
        self.elapsed_regular_time = elapsedRegularTime
        self.match_time = matchTime
        self.type = type
        self.update_id = updateId
        self.update_time = BaseResource.strip_datetime(updateTime)
        self.update_type = updateType
        self.team = team
        self.team_name = teamName
        self.elapsed_added_time = elapsedAddedTime


class EventTimeline(BaseResource):

    def __init__(self, **kwargs):
        super(EventTimeline, self).__init__(**kwargs)
        self.event_id = kwargs.get('eventId')
        self.elapsed_regular_time = kwargs.get('elapsedRegularTime')
        self.event_type_id = kwargs.get('eventTypeId')
        self.in_play_match_status = kwargs.get('inPlayMatchStatus')
        self.status = kwargs.get('status')
        self.time_elapsed = kwargs.get('timeElapsed')
        self.score = Score(**kwargs.get('score'))
        self.update_detail = [UpdateDetail(**i) for i in kwargs.get('updateDetails')]


class FullTimeElapsed(object):

    def __init__(self, hour, min, sec):
        self.hour = hour
        self.min = min
        self.sec = sec


class StateOfBall(object):

    def __init__(self, appealId, appealTypeName, batsmanName, batsmanRuns, bowlerName, bye, dismissalTypeName, legBye,
                 noBall, outcomeId, overBallNumber, overNumber, referralOutcome, wide):
        self.appeal_id = appealId
        self.appeal_type_name = appealTypeName
        self.batsman_name = batsmanName
        self.batsman_runs = batsmanRuns
        self.bowler_name = bowlerName
        self.bye = bye
        self.dismissal_type_name = dismissalTypeName
        self.leg_bye = legBye
        self.no_ball = noBall
        self.outcome_id = outcomeId
        self.over_ball_number = overBallNumber
        self.over_number = overNumber
        self.referral_outcome = referralOutcome
        self.wide = wide


class Scores(BaseResource):

    def __init__(self, **kwargs):
        super(Scores, self).__init__(**kwargs)
        self.event_id = kwargs.get('eventId')
        self.elapsed_regular_time = kwargs.get('elapsedRegularTime')
        self.elapsed_added_time = kwargs.get('elapsedAddedTime')
        self.event_type_id = kwargs.get('eventTypeId')
        self.match_status = kwargs.get('matchStatus')
        self.time_elapsed = kwargs.get('timeElapsed')
        self.time_elapsed_seconds = kwargs.get('timeElapsedSeconds')
        self.status = kwargs.get('status')
        self.current_day = kwargs.get('currentDay')
        self.current_set = kwargs.get('currentSet')
        self.description = kwargs.get('description')
        self.match_type = kwargs.get('matchType')
        self.current_game = kwargs.get('currentGame')
        self.current_point = kwargs.get('currentPoint')
        self.full_time_elapsed = FullTimeElapsed(**kwargs.get('fullTimeElapsed'))
        self.score = Score(**kwargs.get('score'))
        self.state_of_ball = StateOfBall(**kwargs.get('stateOfBall')) if kwargs.get('stateOfBall') else None



import re
import datetime
from requests import ConnectionError

from ..exceptions import APIError, RaceCardError
from ..utils import check_status_code
from .baseendpoint import BaseEndpoint
from .. import resources


class RaceCard(BaseEndpoint):
    """
    RaceCard operations.
    """

    app_key = None

    def login(self, session=None):
        """
        Parses app key from betfair exchange site.

        :param requests.session session: Requests session object
        """
        session = session or self.client.session
        response = session.get(self.login_url)
        app_key = re.findall(r'''"appKey":\s"(.*?)"''', response.text)
        if app_key:
            self.app_key = app_key[0]
        else:
            raise RaceCardError("Unable to find appKey")

    def get_race_card(self, market_ids, data_entries=None, session=None, lightweight=None):
        """
        Returns a list of race cards based on market ids provided.

        :param list market_ids: The filter to select desired markets
        :param str data_entries: Data to be returned
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: list[resources.RaceCard]
        """
        if not self.app_key:
            raise RaceCardError("You need to login before requesting a race_card\n"
                                "APIClient.race_card.login()")
        params = self.create_race_card_req(market_ids, data_entries)
        (response, elapsed_time) = self.request(params=params, session=session)
        return self.process_response(response, resources.RaceCard, elapsed_time, lightweight)

    def request(self, method=None, params=None, session=None):
        session = session or self.client.session
        date_time_sent = datetime.datetime.utcnow()
        try:
            response = session.get(self.url, params=params, headers=self.headers)
        except ConnectionError:
            raise APIError(None, method, params, 'ConnectionError')
        except Exception as e:
            raise APIError(None, method, params, e)
        elapsed_time = (datetime.datetime.utcnow() - date_time_sent).total_seconds()

        response_data = response.json()

        check_status_code(response)
        return response_data, elapsed_time

    @staticmethod
    def create_race_card_req(market_ids, data_entries):
        if not data_entries:
            data_entries = 'RACE, TIMEFORM_DATA, RUNNERS, RUNNER_DETAILS'
        return {
            'dataEntries': data_entries,
            'marketId': ','.join(market_ids)
        }

    @property
    def headers(self):
        return {
            'Connection': 'keep-alive',
            'Content-Type': 'application/json',
            'X-Application': self.app_key
        }

    @property
    def login_url(self):
        return 'https://www.betfair.com/exchange/plus/'

    @property
    def url(self):
        return 'https://www.betfair.com/rest/v2/raceCard'



from .baseresource import BaseResource


class RaceType(object):
    """
    :type abbr: unicode
    :type full: unicode
    :type key: unicode
    """

    def __init__(self, abbr, full, key):
        self.abbr = abbr
        self.full = full
        self.key = key


class RaceClassification(object):
    """
    :type classification: unicode
    :type classification_abbr: unicode
    :type code: unicode
    :type display_name: unicode
    :type display_name_abbr: unicode
    """

    def __init__(self, classification, classificationAbbr, code, displayName, displayNameAbbr):
        self.classification = classification
        self.classification_abbr = classificationAbbr
        self.code = code
        self.display_name = displayName
        self.display_name_abbr = displayNameAbbr


class Market(object):
    """
    :type market_id: unicode
    :type market_type: unicode
    :type number_of_winners: int
    """

    def __init__(self, marketId, marketType, numberOfWinners):
        self.market_id = marketId
        self.market_type = marketType
        self.number_of_winners = numberOfWinners


class Going(object):
    """
    :type abbr: unicode
    :type full: unicode
    :type key: unicode
    """

    def __init__(self, abbr, full, key):
        self.abbr = abbr
        self.full = full
        self.key = key


class Course(object):
    """
    :type country: unicode
    :type country_code: unicode
    :type course_id: unicode
    :type course_type: unicode
    :type name: unicode
    :type surface_type: unicode
    :type timeform_course_code: unicode
    :type timezone: unicode
    """

    def __init__(self, country, countryCode, courseId, name, surfaceType, timeformCourseCode, timezone, courseType=None):
        self.country = country
        self.country_code = countryCode
        self.course_id = courseId
        self.course_type = courseType
        self.name = name
        self.surface_type = surfaceType
        self.timeform_course_code = timeformCourseCode
        self.timezone = timezone


class Race(BaseResource):
    """
    :type betfair_meeting_id: unicode
    :type course: Course
    :type distance: int
    :type eligibility: unicode
    :type going: Going
    :type is_results_available: bool
    :type markets: list[Market]
    :type meeting_going: unicode
    :type meeting_id: unicode
    :type number_of_runners: int
    :type race_class: int
    :type race_classification: RaceClassification
    :type race_id: unicode
    :type race_title: unicode
    :type race_type: RaceType
    :type start_date: datetime.datetime
    """

    def __init__(self, **kwargs):
        super(Race, self).__init__(**kwargs)
        self.betfair_meeting_id = kwargs.get('betfairMeetingId')
        self.distance = kwargs.get('distance')
        self.eligibility = kwargs.get('eligibility')
        self.is_results_available = kwargs.get('isResultAvailable')
        self.meeting_going = kwargs.get('meetingGoing')
        self.meeting_id = kwargs.get('meetingId')
        self.number_of_runners = kwargs.get('numberOfRunners')
        self.race_class = kwargs.get('raceClass')
        self.race_id = kwargs.get('raceId')
        self.race_title = kwargs.get('raceTitle')
        self.start_date = self.strip_datetime(kwargs.get('startDate')) if 'startDate' in kwargs else None
        self.course = Course(**kwargs.get('course')) if 'course' in kwargs else None
        self.going = Going(**kwargs.get('going')) if 'going' in kwargs else None
        self.markets = [Market(**i) for i in kwargs.get('markets') or []]
        self.race_classification = RaceClassification(**kwargs.get('raceClassification')) if 'raceClassification' in kwargs else None
        self.race_type = RaceType(**kwargs.get('raceType')) if 'raceType' in kwargs else None


class InPlayHint(object):
    """
    :type hint_name: unicode
    :type hint_value: unicode
    """

    def __init__(self, hintName, hintValue):
        self.hint_name = hintName
        self.hint_value = hintValue


class DaysSinceLastRun(object):
    """
    :type days: int
    :type type: unicode
    """

    def __init__(self, days, type):
        self.days = days
        self.type = type


class Jockey(object):
    """
    :type jockey_id: unicode
    :type name: unicode
    """

    def __init__(self, jockeyId, name, allowance=None):
        self.jockey_id = jockeyId
        self.name = name
        self.allowance = allowance


class Selection(object):
    """
    :type market_id: unicode
    :type market_type: unicode
    :type selection_id: unicode
    """

    def __init__(self, marketId, marketType, selectionId, bsp=None):
        self.market_id = marketId
        self.market_type = marketType
        self.selection_id = selectionId
        self.bsp = bsp


class Trainer(object):
    """
    :type location: unicode
    :type name: unicode
    :type trainer_id: unicode
    """

    def __init__(self, location, name, trainerId):
        self.location = location
        self.name = name
        self.trainer_id = trainerId


class Wearing(object):
    """
    :type abbr: unicode
    :type full: unicode
    :type key: unicode
    """

    def __init__(self, abbr, full, key):
        self.abbr = abbr
        self.full = full
        self.key = key


class Runner(object):
    """
    :type age: int
    :type comment: unicode
    :type days_since_last_run: DaysSinceLastRun
    :type draw: int
    :type gender: unicode
    :type horse_id: unicode
    :type is_non_runner: bool
    :type jockey: Jockey
    :type long_handicap: int
    :type name: unicode
    :type official_rating: int
    :type owner_colours: unicode
    :type recent_form: unicode
    :type saddle_cloth: unicode
    :type selections: list[Selection]
    :type star_rating: int
    :type timeform_123_place: int
    :type trainer: Trainer
    :type wearing: Wearing
    :type weight: int
    :type wins_at_course: int
    :type wins_at_course_and_distance: int
    :type wins_at_distance: int
    """

    def __init__(self, age=None, draw=None, gender=None, horseId=None, longHandicap=None, name=None, ownerColours=None, saddleCloth=None,
                 weight=None, selections=None, trainer=None, jockey=None, starRating=None, comment=None, isNonRunner=False,
                 winsAtCourse=None, winsAtCourseAndDistance=None, winsAtDistance=None, daysSinceLastRun=None,
                 timeform123Place=None, officialRating=None, recentForm=None, wearing=None):
        self.age = age
        self.comment = comment
        self.draw = draw
        self.gender = gender
        self.horse_id = horseId
        self.is_non_runner = isNonRunner
        self.long_handicap = longHandicap
        self.name = name
        self.official_rating = officialRating
        self.owner_colours = ownerColours
        self.recent_form = recentForm
        self.saddle_cloth = saddleCloth
        self.star_rating = starRating
        self.timeform_123_place = timeform123Place
        self.weight = weight
        self.wins_at_course = winsAtCourse
        self.wins_at_course_and_distance = winsAtCourseAndDistance
        self.wins_at_distance = winsAtDistance
        self.days_since_last_run = [DaysSinceLastRun(**i) for i in daysSinceLastRun] if daysSinceLastRun else []
        self.jockey = Jockey(**jockey) if jockey else None
        self.selections = [Selection(**i) for i in selections] if selections else []
        self.trainer = Trainer(**trainer) if trainer else None
        self.wearing = Wearing(**wearing) if wearing else None


class RaceCard(BaseResource):
    """
    :type betting_forecast_text: unicode
    :type comment: unicode
    :type in_play_hints: InPlayHint
    :type minimum_weight: int
    :type number_of_non_runners: int
    :type number_of_runners: int
    :type prize: unicode
    :type race: Race
    :type runners: list[Runner]
    :type timeform_123_text: unicode
    """

    def __init__(self, **kwargs):
        super(RaceCard, self).__init__(**kwargs)
        self.betting_forecast_text = kwargs.get('bettingForecastText')
        self.comment = kwargs.get('comment')
        self.minimum_weight = kwargs.get('minimumWeight')
        self.number_of_non_runners = kwargs.get('numberOfNonRunners')
        self.number_of_runners = kwargs.get('numberOfRunners')
        self.prize = kwargs.get('prize')
        self.timeform_123_text = kwargs.get('timeform123Text')
        self.in_play_hints = [InPlayHint(**i) for i in
                              kwargs.get('inPlayHints') or []]
        self.race = Race(**kwargs.get('race'))
        self.runners = [Runner(**i) for i in kwargs.get('runners')]



import datetime

from .baseendpoint import BaseEndpoint
from .. import resources
from ..utils import clean_locals


class Scores(BaseEndpoint):
    """
    Scores operations.
    """

    URI = 'ScoresAPING/v1.0/'

    def list_race_details(self, meeting_ids=None, race_ids=None, session=None, lightweight=None):
        """
        Search for races to get their details.

        :param dict meeting_ids: Optionally restricts the results to the specified meeting IDs.
        The unique Id for the meeting equivalent to the eventId for that specific race as
        returned by listEvents
        :param str race_ids: Optionally restricts the results to the specified race IDs. The
        unique Id for the race in the format meetingid.raceTime (hhmm). raceTime is in GMT
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: list[resources.RaceDetail]
        """
        params = clean_locals(locals())
        method = '%s%s' % (self.URI, 'listRaceDetails')
        (response, elapsed_time) = self.request(method, params, session)
        return self.process_response(response, resources.RaceDetails, elapsed_time, lightweight)

    # Following requires app key to be authorised

    def list_available_events(self, event_ids=None, event_type_ids=None, event_status=None, session=None,
                              lightweight=None):
        """
        Search for events that have live score data available.

        :param list event_ids: Optionally restricts the results to the specified event IDs
        :param list event_type_ids: Optionally restricts the results to the specified event type IDs
        :param list event_status: Optionally restricts the results to the specified event status
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: list[resources.AvailableEvent]
        """
        params = clean_locals(locals())
        method = '%s%s' % (self.URI, 'listAvailableEvents')
        (response, elapsed_time) = self.request(method, params, session)
        return self.process_response(response, resources.AvailableEvent, elapsed_time, lightweight)

    def list_scores(self, update_keys, session=None, lightweight=None):
        """
        Returns a list of current scores for the given events.

        :param list update_keys: The filter to select desired markets. All markets that match
        the criteria in the filter are selected e.g. [{'eventId': '28205674', 'lastUpdateSequenceProcessed': 2}]
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: list[resources.Score]
        """
        params = clean_locals(locals())
        method = '%s%s' % (self.URI, 'listScores')
        (response, elapsed_time) = self.request(method, params, session)
        return self.process_response(response, resources.Score, elapsed_time, lightweight)

    def list_incidents(self, update_keys, session=None, lightweight=None):
        """
        Returns a list of incidents for the given events.

        :param dict update_keys: The filter to select desired markets. All markets that match
        the criteria in the filter are selected e.g. [{'eventId': '28205674', 'lastUpdateSequenceProcessed': 2}]
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: list[resources.Incidents]
        """
        params = clean_locals(locals())
        method = '%s%s' % (self.URI, 'listIncidents')
        (response, elapsed_time) = self.request(method, params, session)
        return self.process_response(response, resources.Incidents, elapsed_time, lightweight)

    @property
    def url(self):
        return '%s%s' % (self.client.api_uri, 'scores/json-rpc/v1')



from ..streaming import (
    BaseListener,
    BetfairStream,
)


class Streaming(object):
    """
    Streaming operations.
    """

    def __init__(self, parent):
        """
        :param parent: API client.
        """
        self.client = parent

    def create_stream(self, unique_id, listener=None, timeout=11, buffer_size=1024, description='BetfairSocket'):
        """
        Creates BetfairStream.

        :param dict unique_id: Unique id of the stream
        :param resources.Listener listener:  Listener class to use
        :param float timeout: Socket timeout
        :param int buffer_size: Socket buffer size
        :param str description: Betfair stream description

        :rtype: resources.BetfairStream
        """
        listener = listener if listener else BaseListener()
        return BetfairStream(
            unique_id,
            listener,
            app_key=self.client.app_key,
            session_token=self.client.session_token,
            timeout=timeout,
            buffer_size=buffer_size,
            description=description,
            # todo lightweight
        )



from requests import ConnectionError

from ..exceptions import APIError
from ..utils import check_status_code
from .baseendpoint import BaseEndpoint


class Navigation(BaseEndpoint):
    """
    Navigation operations.
    """

    def list_navigation(self, session=None):
        """
        This Navigation Data for Applications service allows the retrieval of the
        full Betfair market navigation menu from a compressed file.

        :param requests.session session: Requests session object

        :rtype: json
        """
        return self.request(session=session)

    def request(self, method=None, params=None, session=None):
        session = session or self.client.session
        try:
            response = session.get(self.url, headers=self.client.request_headers,
                                   timeout=(self.connect_timeout, self.read_timeout))
        except ConnectionError:
            raise APIError(None, method, params, 'ConnectionError')
        except Exception as e:
            raise APIError(None, method, params, e)

        check_status_code(response)
        return response.json()

    @property
    def url(self):
        return self.client.navigation_uri



import ujson as json
import threading
import socket
import ssl
import datetime

from ..exceptions import SocketError
from ..compat import is_py3


class BetfairStream(object):
    """Socket holder, connects to betfair and
    pushes any received data to listener
    """

    __host = 'stream-api.betfair.com'
    __port = 443
    __CRLF = '\r\n'
    __encoding = 'utf-8'

    def __init__(self, unique_id, listener, app_key, session_token, timeout, buffer_size, description):
        self.unique_id = unique_id
        self.listener = listener
        self.app_key = app_key
        self.session_token = session_token
        self.timeout = timeout
        self.buffer_size = buffer_size
        self.description = description
        self.receive_count = 0
        self.datetime_last_received = None

        self._socket = None
        self._running = False

    def start(self, async=False):
        """Starts read loop, new thread if async and
        connects/authenticates if not already running.

        :param async: If True new thread is started
        """
        if not self._running:
            self._connect()
            self.authenticate()
        if async:
            t = threading.Thread(name=self.description, target=self._read_loop)
            t.daemon = False
            t.start()
        else:
            self._read_loop()

    def stop(self):
        """Stops read loop and closes socket if it has been created.
        """
        self._running = False

        if self._socket is not None:
            try:
                self._socket.shutdown(2)  # SHUT_RDWR
            except OSError:
                pass

    def authenticate(self, unique_id=None):
        """Authentication request.

        :param unique_id: self.unique_id used if not supplied.
        """
        message = {
            'op': 'authentication',
            'id': unique_id or self.unique_id,
            'appKey': self.app_key,
            'session': self.session_token
        }
        self.listener.register_stream(unique_id or self.unique_id, 'authentication')
        self._send(message)

    def heartbeat(self, unique_id=None):
        """Heartbeat request to keep session alive.

        :param unique_id: self.unique_id used if not supplied.
        """
        message = {
            'op': 'heartbeat',
            'id': unique_id or self.unique_id
        }
        self._send(message)

    def subscribe_to_markets(self, unique_id, market_filter, market_data_filter, initial_clk=None, clk=None):
        """
        Market subscription request.

        :param dict market_filter: Market filter
        :param dict market_data_filter: Market data filter
        :param int unique_id: Unique id of stream
        :param str initial_clk: Sequence token for reconnect
        :param str clk: Sequence token for reconnect
        """
        message = {
            'op': 'marketSubscription',
            'id': unique_id,
            'marketFilter': market_filter,
            'marketDataFilter': market_data_filter,
            'initialClk': initial_clk,
            'clk': clk,
        }
        self.listener.register_stream(unique_id, 'marketSubscription')
        self._send(message)

    def subscribe_to_orders(self, unique_id, order_filter=None, initial_clk=None, clk=None):
        """
        Order subscription request.

        :param int unique_id: Unique id of stream
        :param dict order_filter: Order filter to be applied
        :param str initial_clk: Sequence token for reconnect
        :param str clk: Sequence token for reconnect
        """
        message = {
            'op': 'orderSubscription',
            'id': unique_id,
            'orderFilter': order_filter,
            'initialClk': initial_clk,
            'clk': clk,
        }
        self.listener.register_stream(unique_id, 'orderSubscription')
        self._send(message)

    def _connect(self):
        """Creates socket and sets running to True.
        """
        self._socket = self._create_socket()
        self._running = True

    def _create_socket(self):
        """Creates ssl socket, connects to stream api and
        sets timeout.
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s = ssl.wrap_socket(s)
        s.connect((self.__host, self.__port))
        s.settimeout(self.timeout)
        return s

    def _read_loop(self):
        """Read loop, splits by CRLF and pushes received data
        to _data.
        """
        while self._running:
            try:
                received_data_raw = self._receive_all()
                if self._running:
                    self.receive_count += 1
                    self.datetime_last_received = datetime.datetime.utcnow()
                    received_data_split = received_data_raw.split(self.__CRLF)
                    for received_data in received_data_split:
                        if received_data:
                            self._data(received_data)
            except (socket.timeout, socket.error) as e:
                self.stop()
                raise SocketError('[Connect: %s]: Socket %s' % (self.unique_id, e))

    def _receive_all(self):
        """Whilst socket is running receives data from socket,
        till CRLF is detected.
        """
        (data, part) = ('', '')
        if is_py3:
            crlf_bytes = bytes(self.__CRLF, encoding=self.__encoding)
        else:
            crlf_bytes = self.__CRLF

        while self._running and part[-2:] != crlf_bytes:
            part = self._socket.recv(self.buffer_size)
            if part:
                data += part.decode(self.__encoding)
        return data

    def _data(self, received_data):
        """Sends data to listener, if False is returned; socket
        is closed.

        :param received_data: Decoded data received from socket.
        """
        if self.listener.on_data(received_data) is False:
            self.stop()

    def _send(self, message):
        """If not running connects socket and
        authenticates. Adds CRLF and sends message
        to Betfair.

        :param message: Data to be sent to Betfair.
        """
        if not self._running:
            self._connect()
            self.authenticate()
        message_dumped = json.dumps(message) + self.__CRLF
        self._socket.send(message_dumped.encode())

    def __str__(self):
        return '<BetfairStream [%s]>' % ('running' if self._running else 'not running')

    def __repr__(self):
        return '<BetfairStream>'



import datetime
import logging
import time

from ..resources.streamingresources import (
    MarketBookCache,
    OrderBookCache,
)

logger = logging.getLogger(__name__)


class BaseStream(object):
    """Separate stream class to hold market/order caches
    """

    _lookup = 'mc'

    def __init__(self, unique_id, output_queue, max_latency, lightweight):
        self.unique_id = unique_id
        self.output_queue = output_queue
        self._max_latency = max_latency
        self._lightweight = lightweight

        self._initial_clk = None
        self._clk = None
        self._caches = {}
        self._updates_processed = 0
        self._on_creation()

        self.time_created = datetime.datetime.utcnow()
        self.time_updated = datetime.datetime.utcnow()

    def on_subscribe(self, data):
        self._update_clk(data)
        publish_time = data.get('pt')

        book_data = data.get(self._lookup)
        if book_data:
            self._process(book_data, publish_time)
        logger.info('[Stream: %s]: %s %s added' % (self.unique_id, len(self._caches), self._lookup))

    def on_heartbeat(self, data):
        self._update_clk(data)

    def on_resubscribe(self, data):
        self._update_clk(data)

    def on_update(self, data):
        self._update_clk(data)

        publish_time = data.get('pt')
        latency = self._calc_latency(publish_time)
        if latency > self._max_latency:
            logger.warning('[Stream: %s]: Latency high: %s' % (self.unique_id, latency))

        book_data = data.get(self._lookup)
        self._process(book_data, publish_time)

    def clear_cache(self):
        self._caches.clear()

    def _on_creation(self):
        logger.info('[Stream: %s]: "%s" created' % (self.unique_id, self))

    def _process(self, book_data, publish_time):
        pass

    def _update_clk(self, data):
        (initial_clk, clk) = (data.get('initialClk'), data.get('clk'))
        if initial_clk:
            self._initial_clk = initial_clk
        if clk:
            self._clk = clk
        self.time_updated = datetime.datetime.utcnow()

    @staticmethod
    def _calc_latency(publish_time):
        return time.time() - publish_time / 1e3

    def __str__(self):
        return '<BaseStream>'

    def __repr__(self):
        return str(self)


class MarketStream(BaseStream):

    _lookup = 'mc'

    def _process(self, market_books, publish_time):
        output_market_book = []
        for market_book in market_books:
            market_id = market_book.get('id')
            if market_book.get('img'):
                self._caches[market_id] = MarketBookCache(publish_time=publish_time, **market_book)
                logger.info('[MarketStream: %s] %s added' % (self.unique_id, market_id))
            else:
                market_book_cache = self._caches.get(market_id)
                if market_book_cache:
                    market_book_cache.update_cache(market_book, publish_time)
                    self._updates_processed += 1
                else:
                    logger.error('[MarketStream: %s] Received update for market not in cache: %s' %
                                 (self.unique_id, market_book))
            output_market_book.append(
                self._caches[market_id].create_market_book(self.unique_id, market_book, self._lightweight)
            )

        self.output_queue.put(output_market_book)

    def __str__(self):
        return 'MarketStream'

    def __repr__(self):
        return '<MarketStream [%s]>' % len(self._caches)


class OrderStream(BaseStream):

    _lookup = 'oc'

    def _process(self, order_books, publish_time):
        output_order_book = []
        for order_book in order_books:
            market_id = order_book.get('id')
            order_book_cache = self._caches.get(market_id)
            if order_book_cache:
                order_book_cache.update_cache(order_book, publish_time)
            else:
                self._caches[market_id] = OrderBookCache(**order_book)
                logger.info('[OrderStream: %s] %s added' % (self.unique_id, market_id))
            self._updates_processed += 1

            output_order_book.append(
                self._caches[market_id].create_order_book(self.unique_id, order_book, self._lightweight)
            )

        self.output_queue.put(output_order_book)

    def __str__(self):
        return 'OrderStream'

    def __repr__(self):
        return '<OrderStream [%s]>' % len(self._caches)



from .baseclient import BaseClient
from . import endpoints


class APIClient(BaseClient):

    def __init__(self, username, password=None, app_key=None, certs=None, locale=None, cert_files=None,
                 lightweight=False):
        """
        Creates API client for API operations.

        :param str username: Betfair username
        :param str password: Password for supplied username, if None will look in .bashprofile
        :param str app_key: App Key for account, if None will look in .bashprofile
        :param str certs: Directory for certificates, if None will look in /certs/
        :param str locale: Exchange to be used, defaults to UK for login and global for api
        :param list cert_files: Certificate and key files. If None will look in `certs`
        :param bool lightweight: If True endpoints will return dict not a resource (22x faster)
        """
        super(APIClient, self).__init__(
            username, password, app_key=app_key, certs=certs, locale=locale, cert_files=cert_files,
            lightweight=lightweight
        )

        self.login = endpoints.Login(self)
        self.keep_alive = endpoints.KeepAlive(self)
        self.logout = endpoints.Logout(self)
        self.betting = endpoints.Betting(self)
        self.account = endpoints.Account(self)
        self.navigation = endpoints.Navigation(self)
        self.scores = endpoints.Scores(self)
        self.streaming = endpoints.Streaming(self)
        self.in_play_service = endpoints.InPlayService(self)
        self.race_card = endpoints.RaceCard(self)

    def __repr__(self):
        return '<APIClient [%s]>' % self.username

    def __str__(self):
        return 'APIClient'
